dnl @synopsis Checks if OpenSSL library is available.
dnl
dnl This macro will check various standard spots for OpenSSL including
dnl a user-supplied directory.
dnl The user uses '--with-ssl' or '--with-ssl=/path/to/ssl' as arguments
dnl to configure.
dnl
dnl Unlike Mark's original code, this macro will always check for
dnl OpenSSL and will force it's presence.
dnl
dnl If OpenSSL is found the include directory gets added to CFLAGS,
dnl '-lssl' & '-lcrypto' get added to LIBS, and the libraries
dnl location gets added to LDFLAGS.
dnl
dnl @version $Id: acinclude.m4,v 1.3 2002/07/15 11:12:54 fedy Exp $
dnl @author Jan Fedak <jack@mobil.cz>
dnl Based on code by Mark Ethan Trostler <trostler@juniper.net>
dnl
AC_DEFUN([CHECK_SSL],[
    AC_MSG_CHECKING(for OpenSSL)
    AC_ARG_WITH(open-ssl,
    [[  --with-open-ssl[=DIR]   specify path for OpenSSL [will check /usr/local/ssl
                               /usr/lib/ssl /usr/ssl /usr/pkg /usr/local /usr]]],
    [
	if test "$withval" = "no"; then
	    AC_MSG_RESULT(no)
	    AC_MSG_ERROR(OpenSSL is required.)
	else
	    if test "$withval" = "yes"; then
		withval=""
	    fi
	fi
    ])
    for dir in $withval /usr/local/ssl /usr/lib/ssl /usr/ssl /usr/pkg /usr/local /usr; do
	ssldir="$dir"
	if test -f "$dir/include/openssl/ssl.h"; then
	    found_ssl="yes";
	    CFLAGS="$CFLAGS -I$ssldir/include/openssl -I$ssldir/include";
	    break;
	fi
	if test -f "$dir/include/ssl.h"; then
	    found_ssl="yes";
	    CFLAGS="$CFLAGS -I$ssldir/include";
	    break
	fi
    done
    if test x$found_ssl != xyes; then
	AC_MSG_ERROR(Cannot find OpenSSL libraries)
    else
	AC_MSG_RESULT(found in $ssldir)
    fi
    LIBS="$LIBS -lssl -lcrypto";
    LDFLAGS="$LDFLAGS -L$ssldir/lib";
])

dnl @synopsis Select method of PRNG initialization.
dnl
dnl This macro will check various standard spots for OpenSSL including
dnl a user-supplied directory.
dnl The user uses '--with-prng-init=method' as arguments to configure.
dnl
dnl @version $Id: acinclude.m4,v 1.3 2002/07/15 11:12:54 fedy Exp $
dnl @author Jan Fedak <jack@mobil.cz>
dnl
AC_DEFUN([SELECT_PRNG_INIT],[
    AC_MSG_CHECKING(PRNG init method)
    AC_ARG_WITH(prng-init,
	AC_HELP_STRING([--with-prng-init=METHOD],
		       [select method of PRNG initialization. Methods are
		       auto, urandom, windows, prngd, randfile, pseudo.]),
    [
	# First part -- we check the user input
	if test "x$withval" = "xno"; then
	    AC_MSG_RESULT(no)
	    AC_MSG_ERROR(PRNG initialization is required.)
	fi
	if test "x$withval" = "xauto"; then
	    prng_method=auto
	elif test "x$withval" = "xurandom"; then
	    prng_method=urandom
	elif test "x$withval" = "xwindows"; then
	    prng_method=windows
	elif test "x$withval" = "xprngd"; then
	    prng_method=prngd
	elif test "x$withval" = "xrandfile"; then
	    prng_method=randfile
	elif test "x$withval" = "xpseudo"; then
	    prng_method=pseudo
	else
	    AC_MSG_RESULT(unknown method)
	    AC_MSG_ERROR(PRNG initialization is required.)
	fi
    ],
    [
	prng_method=auto
    ])
    # Second part -- we select method if auto was given
    if test "x$prng_method" = "xauto"; then
	# TODO: test windoze
	if test -r /dev/urandom; then
	    prng_method=urandom
	elif test -S /var/spool/prngd/pool; then
	    prng_method=prngd
	    prngd_socket=/var/spool/prngd/pool
	else
	    prng_method=randfile
	fi
    fi
    # Third part -- we check and process the selection
    if test "x$prng_method" = "xurandom"; then
	AC_MSG_RESULT(urandom)
	AC_DEFINE(PRNG_USE_DEV_URANDOM, 1, [Use /dev/urandom for PRNG initialization.])
	if test ! -r /dev/urandom; then
	    AC_MSG_ERROR(/dev/urandom not found)
	fi
	prng_method=urandom
    elif test "x$prng_method" = "xwindows"; then
	# TODO: check that Windows are present
	AC_MSG_RESULT(windows)
	prng_method=windows
    elif test "x$prng_method" = "xprngd"; then
	AC_MSG_RESULT(prngd)
	AC_DEFINE(PRNG_USE_PRNGD, 1, [Use PRNGD or EGD daemon for PRNG initialization.])
	prng_method=prngd
    elif test "x$prng_method" = "xrandfile"; then
	AC_MSG_RESULT(randfile)
	AC_DEFINE(PRNG_USE_RANDFILE, 1, [Use randfile for PRNG initialization.])
	AC_MSG_WARN([[Make sure your rand file contains enough entropy!]])
	prng_method=randfile
    elif test "x$prng_method" = "xpseudo"; then
	AC_MSG_RESULT(pseudo)
	AC_DEFINE(PRNG_USE_PSEUDO, 1, [Use pseudorandom generator for PRNG initialization.])
	AC_MSG_WARN([[PRNG initialization with pseudo-random bytes is insecure!]])
	prng_method=pseudo
    fi

    AC_ARG_WITH(prngd-sock,
	AC_HELP_STRING([--with-prngd-sock=<PATH>],
		       [Give path to PRNGD or EGD socket (required if --with-prng-init=prngd)]),
	[
	    if test "x$withval" != "xno"; then
		if test "x$prng_method" != "xprngd"; then
		    AC_MSG_ERROR([[PRNGD or EGD sock given but PRNG initialization method is not prngd.]])
		fi
		prngd_socket=$withval
	    else
		AC_MSG_ERROR([[PRNGD or EGD sock not given. Use --with-prngd-sock=<PATH>]])
	    fi
	])
    if test "x$prng_method" = "xprngd"; then
	if test "x$prngd_socket" = "x"; then
	    # we try to guess prngd or prngd socket path
	    if test -S /var/spool/prngd/pool; then
		prngd_socket=$withval
	    else
		AC_MSG_ERROR([[PRNGD or EGD sock not given. Use --with-prngd-sock=<PATH>]])
	    fi
	fi
	AC_DEFINE_UNQUOTED(PRNGD_SOCKET_PATH, "$prngd_socket", [Path to PRNGD or EGD socket.])
	AC_MSG_NOTICE([[Path to PRNGD socket is $prngd_socket]])
    fi

    AC_ARG_WITH(randfile,
	AC_HELP_STRING([--with-randfile=<PATH>],
		       [Give path to randfile (for --with-prng-init=randfile)]),
	[
	    if test "x$withval" != "xno"; then
		if test "x$prng_method" != "xrandfile"; then
		    AC_MSG_ERROR([[Path to randfile given but PRNG initialization method is not randfile.]])
		fi
		AC_DEFINE_UNQUOTED(RANDFILE, "$withval", [Path to randfile.])
		AC_MSG_NOTICE([[Path to randfile is $withval]])
	    fi
	])
])
