/**************************************************************************\
ccvssh - a secure bridge for cvs pservers
Copyright (C) 2002 Jan Fedak
Copyright (C) 2002 Sabren Enterprises, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
02111-1307, USA.
-------------------------------------------------------------
ccvssh - secure ext-to-pserver bridge for cvs

This program illustrates a new (and hopefully superior)
method of securing a CVS pserver.

It offers  an ext-to-pserver bridge: From the client CVS's
point of view, we look like ssh, but what we really do is
make our own ssl connection to a stunnel demon. This gives
us the security and client-side ease of ext/ssh with the
flexibility of pserver on the repository.

Originally written in Python by Michael J. Wallace. 
See http://sabren.net/code/cvssh/ for more information and 
great pictures :)

Modified and rewritten to C by Jan Fedak (jack@mobil.cz)


TODO:
    - proper autoconf configuration (currently compiles only on my system :)
    - support for older .cvspass format
    - remove various TOFIXes
    - optionally check server certificates
    - ensure that it compiles (and runs :) on Windows -- there'll be
      definitelly problem with the select() call
\**************************************************************************/

#define SCVS_PORT		2405
#define LINE_LEN 		1024
#define BUFFER_SIZE		2048
#define CVS_PASSWORD_FILE	".cvspass"

#undef DEBUG
#define MIN_PRNG_SEED_SIZE	64

#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <pwd.h>

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <ssl.h>
#include <err.h>
#include <rand.h>

#include "config.h"
#include "scramble.h"

SSL_CTX *ssl_ctx = NULL;
SSL *ssl = NULL;
X509 *cert = NULL;
#if defined PRNG_USE_RANDFILE
int allow_save_prng_entropy = 0;
#endif


/* MEMORY ALLOCATION ROUTINES {{{ */
/* xstrdup() {{{
 * strdup + additional check
 */
char *xstrdup(char const *str)
{
    char *retval = strdup(str);
    if (!retval)
    {
	fprintf(stderr, "Insufficient memory.\n");
	exit(1);
    }
    return retval;
}
/* }}} */


/* xmalloc() {{{
 * malloc + additional check
 */
void *xmalloc(size_t size)
{
    void *retval = malloc(size);
    if (!retval)
    {
	fprintf(stderr, "Insufficient memory.\n");
	exit(1);
    }
    return retval;
}
/* }}} */
/* }}} */


/* SSL ROUTINES {{{ */
/* PRNG ROUTINES {{{ */
#if defined PRNG_USE_RANDFILE
/* get_rand_file_name() {{{
 * Will return path to the file with entropy
 */
char *get_rand_file_name()
{
# if defined RANDFILE
    /* either given at 'configure' time... {{{ */
    static char *rand_file_name = RANDFILE;
    /* }}} */
# else
    /* or the default one ($RANDFILE or $HOME/.rnd) {{{ */
    static char rand_file_name[FILENAME_MAX];
    if (!RAND_file_name(rand_file_name, sizeof(rand_file_name)))
    {
	fprintf(stderr,
		"Unable to get 'rand' file with entropy.\n"
		"Set the RANDFILE environment variable and create the file with:\n"
		"  openssl rand -out $RANDFILE 1024\n"
		"For details consult the OpenSSL documentation.\n");
	exit(1);
    }
    /* }}} */
# endif
    return rand_file_name;
}
/* }}} */
#endif


/* init_prng() {{{ 
 * Initializes pseudo-random number generator of SSL. */
void init_prng()
{
#if defined WINDOWS
    /* on Windoze we initialize PRNG with the content of the screen {{{ */
    RAND_screen();
    /* }}} */
#elif defined PRNG_USE_DEV_URANDOM
    /* we have /dev/urandom, no need to do anything... {{{ */
    /* }}} */
#elif defined PRNG_USE_PRNGD
    /* we will use PRNGD or EGD daemon {{{ */
    /* 64B (= 512) bytes of entropy should suffice... */
    if (RAND_egd_bytes(PRNGD_SOCKET_PATH, MIN_PRNG_SEED_SIZE) != MIN_PRNG_SEED_SIZE)
    {
	fprintf(stderr,
		"Unable to read enough entropy from the socket `%s'.\n",
		PRNGD_SOCKET_PATH);
	exit(1);
    }
    /* }}} */
#elif defined PRNG_USE_PSEUDO
    /* we will initialize PRNG with pseudo random bytes {{{ */
# warning PRNG will be seeded with pseudo random bytes!
    char prnd_bytes[1024];
    RAND_pseudo_bytes(prnd_bytes, sizeof(prnd_bytes));
    RAND_seed(prnd_bytes, sizeof(prnd_bytes));
    /* }}} */
#elif defined PRNG_USE_RANDFILE
    /* we will read 'random' bytes from a file {{{ */
    int numbytes;
    char *rand_file_name = get_rand_file_name();
    /* 64B (= 512) bytes of entropy should suffice... */
    numbytes = RAND_load_file(rand_file_name, -1);
    if (numbytes < MIN_PRNG_SEED_SIZE)
    {
	fprintf(stderr,
		"Unable to read enough entropy from `%s'.\n",
		rand_file_name);
	exit(1);
    }
    if (numbytes > 512)
	allow_save_prng_entropy = 1;
    /* }}} */
#else
    #error No PRNG seeding method defined
#endif
}
/* }}} */


/* save_prng_entropy() {{{
 * Tries to save entropy back to the RANDFILE.
 */
void save_prng_entropy()
{
#if defined PRNG_USE_RANDFILE
    if (allow_save_prng_entropy)
    {
	char *rand_file_name = get_rand_file_name();
	RAND_write_file(rand_file_name);
    }
#endif
}
/* }}} */
/* }}} */


/* init_ssl() {{{
 * Will initialize the SSL routines and context.
 */
void init_ssl()
{
    OpenSSL_add_ssl_algorithms();
    SSL_load_error_strings();

    if (!(ssl_ctx=SSL_CTX_new(SSLv3_client_method())))
    {
	ERR_print_errors_fp(stderr);
	exit(1);
    }

    init_prng();
}
/* }}} */


/* shut_ssl() {{{
 */
void shut_ssl() 
{
    save_prng_entropy();
    if (ssl_ctx)
    {
	SSL_CTX_free(ssl_ctx);
	ssl_ctx = NULL;
    }
}
/* }}} */


/* open_ssl_connection() {{{
 * Will connect to the given server.
 */
void open_ssl_connection(char *server, int port)
{
    struct sockaddr_in sock_addr;
    struct hostent *hent;
    int sock;

    /* The usual socket stuff... */
    if (!(hent = gethostbyname(server)))
    {
	fprintf(stderr, "Cannot resolve hostname `%s' (%s)\n", server, strerror(errno));
	exit(1);
    }
    
    if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1)
    {
	fprintf(stderr, "Cannot create socket (%s)\n", strerror(errno));
	exit(1);
    }

    sock_addr.sin_family = hent->h_addrtype;
    memcpy(&sock_addr.sin_addr.s_addr, hent->h_addr, hent->h_length);
    sock_addr.sin_port = htons(port);

    if (connect(sock, (struct sockaddr*)&sock_addr, sizeof(struct sockaddr_in)) == -1)
    {
	fprintf(stderr, "Cannot connect to `%s:%d' (%s)\n", server, port, strerror(errno));
	exit(1);
    }

    /* and now some SSL */
    if (!(ssl = SSL_new(ssl_ctx)))
    {
	ERR_print_errors_fp(stderr);
	exit(1);
    }
    if (!SSL_set_fd(ssl, sock))
    {
	ERR_print_errors_fp(stderr);
	exit(1);
    }
    if (SSL_connect(ssl) != 1)
    {
	ERR_print_errors_fp(stderr);
	exit(1);
    }

    /* let's ask for certificate, just for fun... */
    if (!(cert = SSL_get_peer_certificate(ssl)))
    {
	fprintf(stderr, "Cannot get server certificate!\n");
	ERR_print_errors_fp(stderr);
	exit(1);
    } 
}
/* }}} */


/* close_ssl_connection() {{{
 * Will close open connection.
 */
void close_ssl_connection()
{
    if (ssl)
    {
	int sock = SSL_get_fd(ssl);
	SSL_free(ssl);
	close(sock);
	ssl = NULL;
    }
    
    if (cert)
    {
	X509_free(cert);
	cert = NULL;
    }
}
/* }}} */


/* send_via_ssl() {{{
 * Guess what it does :)
 */
void send_via_ssl(char *buffer, int len)
{
    if (!len)
	len = strlen(buffer);

    if (SSL_write(ssl, buffer, len) != len)
    {
	ERR_print_errors_fp(stderr);
	exit(1);
    }
}
/* }}} */


/* forward_stdinout_to_ssl_connection() {{{
 * Will send all the stuff to and fro...
 */
void forward_stdinout_to_ssl_connection()
{
    char buffer[BUFFER_SIZE];

    for (;;)
    {
	fd_set rfds, efds;
	int sock = SSL_get_fd(ssl);
	int retval, count;
	
	if (SSL_pending(ssl))
	{
	    retval = 1;
	    #ifdef DEBUG
		fprintf(stderr, "\n skipping select()\n");
	    #endif
	}
	else
	{
	    FD_ZERO(&rfds);
	    FD_ZERO(&efds);
	    FD_SET(0, &rfds);
	    FD_SET(0, &efds);
	    FD_SET(sock, &rfds);
	    FD_SET(sock, &efds);
	    
	    #ifdef DEBUG
		fprintf(stderr, "\n before select()\n");
	    #endif
	    retval = select(sock+1, &rfds, NULL, &efds, NULL);
	    #ifdef DEBUG
		fprintf(stderr, "\n after select(): %d\n", retval);
	    #endif
	}
	if (retval)
	{
	    /* copy stdin => server {{{ */
	    if (FD_ISSET(0, &rfds))
	    {
		if ((count = read(0, buffer, BUFFER_SIZE)) == 0)
		{
		    /* the client closed the connection. our job's over */
		    return;
		}
		if (SSL_write(ssl, buffer, count) != count)
		{
		    ERR_print_errors_fp(stderr);
		    return;
		}
		#ifdef DEBUG
		    fprintf(stderr, "\n\n>>> ");
		    fwrite(buffer, 1, count, stderr);
		#endif
	    }
	    /* }}} */
	    /* copy server => stdout {{{ */
	    if (FD_ISSET(sock, &rfds))
	    {
		if ((count = SSL_read(ssl, buffer, BUFFER_SIZE)) <= 0)
		{
		    ERR_print_errors_fp(stderr);
		    return;
		}
		if (write(1, buffer, count) != count)
		{
		    fprintf(stderr, "Writing to stdout failed!\n");
		    return;
		}
		#ifdef DEBUG
		    fprintf(stderr, "\n\n<<< ");
		    fwrite(buffer, 1, count, stderr);
		#endif
	    }
	    /* }}} */
	}
	if (FD_ISSET(sock, &efds) ||
	    FD_ISSET(0, &efds))
	{
	    fprintf(stderr, "Exception!\n");
	    return;
	}
    }
}
/* }}} */
/* }}} */


/* CVS SPECIFIC FUNCTIONS {{{ */
/* construct_cvspass_filename() {{{
 * Taken almost 1:1 from CVS.
 * The return value will need to be freed.
 */
char *construct_cvspass_filename ()
{
    char *homedir;
    char *passfile;

    /* Environment should override file. */
    if ((passfile = getenv ("CVS_PASSFILE")) != NULL)
	return xstrdup(passfile);

    /* Construct absolute pathname to user's password file. */
    homedir = getenv("HOME");
    if (!homedir)
    {
	/* FIXME?  This message confuses a lot of users, at least
	   on Win95 (which doesn't set HOMEDRIVE and HOMEPATH like
	   NT does).  I suppose the answer for Win95 is to store the
	   passwords in the registry or something (??).  And .cvsrc
	   and such too?  Wonder what WinCVS does (about .cvsrc, the
	   right thing for a GUI is to just store the password in
	   memory only)...  */
	fprintf(stderr, "Could not find out home directory.\n");
	exit(1);
    }

    passfile = (char *)xmalloc(strlen(homedir) + strlen(CVS_PASSWORD_FILE) + 3);
    strcpy (passfile, homedir);
    strcat (passfile, "/");
    strcat (passfile, CVS_PASSWORD_FILE);

    /* TOFIX: awful laziness has hit my fingers :( ...
    if (isfile(passfile))
	chmod(passfile, 0600);
    */

    return passfile;
}
/* }}} */


/* extract_cvsroot_info() {{{
 * Will extract info about user and server from CVSROOT. Will allocate memory for string storage.
 * Returns 0 on success, anything else on failure.
 * "[:ext:][[user]@]host[:]/path",
 */
int extract_cvsroot_info(char **username, char **password, char **server, int *port, char **root)
{
    char *cvsroot = getenv("CVSROOT");
    char *cvsroot_save = NULL;		/* what we allocate */
    char *cvsroot_copy = NULL;		/* work copy (will write over it) */
    char *p, *q;			/* help variables */
    char *firstslash;			/* save where the path spec starts */
    
    /* First we clear all result values. */
    *username = *password = *server = *root = 0;
    *port = SCVS_PORT;
    
    /* Is the environment variable set? {{{ */
    if (!cvsroot)
    {
	fprintf(stderr, "CVSROOT environment variable is not set.\n");
	goto error_exit;
    }
    /* }}} */
    /* Make copy {{{ */
    cvsroot_save = cvsroot_copy = xstrdup(cvsroot);
    /* }}} */
    /* Method specifier. {{{ */
    if (*cvsroot_copy == ':')
    {
	if (strncmp(cvsroot_copy, ":ext:", 5))
	{
	    fprintf(stderr, "CVSROOT: invalid method specified. Use :ext:\n");
	    fprintf(stderr, "CVSROOT is `%s'", cvsroot);
	    goto error_exit;
	}
	cvsroot_copy += 5;
    }
    /* }}} */
    /* Looking for path {{{ */
    if ((p = strchr (cvsroot_copy, '/')) == NULL)
    {
	fprintf(stderr, "CVSROOT requires a path spec:\n[:ext:][[user]@]host[:]/path\n");
	fprintf(stderr, "CVSROOT is `%s'", cvsroot);
	goto error_exit;
    }
    firstslash = p;
    *p = '\0';
    /* }}} */

    /* Check to see if there is a username[:password] in the string. {{{ */
    if ((p = strchr (cvsroot_copy, '@')) != NULL)
    {
	*p = '\0';
	/* Check for a password {{{ */
	if ((q = strchr(cvsroot_copy, ':')) != NULL)
	{
	    *q = '\0';
	    *password = xstrdup(++q);
	    /* "" means empty password, NULL means password from .cvspass */
	}
	/* }}} */

	/* Copy the username {{{ */
	if (*cvsroot_copy != '\0')
	    *username = xstrdup(cvsroot_copy);
	else
	{
	    /* Blank name, we use current user. */
	    struct passwd *pwd = getpwuid(getuid());
	    if (!pwd)
	    {
		fprintf(stderr, "Cannot get user name for uid `%d'.\n", getuid());
		goto error_exit;
	    }
	    *username = xstrdup(pwd->pw_name);
	}
	/* }}} */

	cvsroot_copy = ++p;
    }
    /* }}} */

    /* Now deal with host[:[port]] {{{ */
    /* The port {{{ */
    if ((p = strchr (cvsroot_copy, ':')) != NULL)
    {
	*p++ = '\0';
	if (strlen(p))
	{
	    q = p;
	    while (*q)
	    {
		if (!isdigit(*q++))
		{
		    fprintf(stderr, "CVSROOT may only specify a positive, non-zero, integer port (not `%s').\n", p);
		    fprintf(stderr, "Perhaps you entered a relative pathname?\n");
		    goto error_exit;
		}
	    }
	    if ((*port = atoi(p)) <= 0)
	    {
		fprintf(stderr, "CVSROOT may only specify a positive, non-zero, integer port (not `%s').\n", p);
		fprintf(stderr, "Perhaps you entered a relative pathname?\n");
		goto error_exit;
	    }
	}
    }
    /* }}} */
    /* Copy host {{{ */
    if (*cvsroot_copy == '\0')
    {
	fprintf(stderr, "CVSROOT doesn't contain server address.\n");
	goto error_exit;
    }
    *server = xstrdup(cvsroot_copy);
    /* }}} */
    /* }}} */

    /* restore the '/' */
    cvsroot_copy = firstslash;
    *cvsroot_copy = '/';

    *root = xstrdup(cvsroot_copy);

    free(cvsroot_save);

    /* ok, exit */
    return 0;

error_exit:
    /* free memory and leave */
    if (cvsroot_save)
	free(cvsroot_save);
    #define FREE_RESULT_POINTER(ptr)	if (*ptr) { free(*ptr); *ptr = NULL; }
    FREE_RESULT_POINTER(username);
    FREE_RESULT_POINTER(password);
    FREE_RESULT_POINTER(server);
    FREE_RESULT_POINTER(root);
    return -1;
}
/* }}} */


/* get_cannonical_cvsroot() {{{
 * Normalizes CVSROOT to the shape required in .cvspass
 * Allocates and returns new string.
 */
char *get_cannonical_cvsroot(char *username, char *server, int port, char *root)
{
    char *lowercase_server = xstrdup(server);	/* server name in lowercase */
    char port_s[32];				/* port as string */
    char *cannonical_cvsroot = NULL;		/* result */
    char *c;					/* tmp var */
    
    for (c = lowercase_server; *c; c++)
	*c = tolower(*c);
    sprintf(port_s, "%d", port);

    cannonical_cvsroot = xmalloc(strlen(username) +
				 strlen(lowercase_server) + 
				 strlen(port_s) +
				 strlen(root) + 12);
    sprintf(cannonical_cvsroot, ":pserver:%s@%s:%s%s",
	    username, lowercase_server, port_s, root);

    free(lowercase_server);

    return cannonical_cvsroot;
}
/* }}} */


/* validate_password_on_cvs_server() {{{
 * Will validate the password on the server.
 */
int validate_password_on_cvs_server(char *username, char *password, char *root)
{
    char *scrambled_password = scramble(password);
    #define REPLY_LEN 32
    char reply[REPLY_LEN + 1];
    char correct_reply[] = "I LOVE YOU\n";
    int len;

    send_via_ssl("BEGIN AUTH REQUEST\n", 0);
    send_via_ssl(root, 0);
    send_via_ssl("\n", 0);
    send_via_ssl(username, 0);
    send_via_ssl("\n", 0);
    send_via_ssl(scrambled_password, 0);
    send_via_ssl("\n", 0);
    send_via_ssl("END AUTH REQUEST\n", 0);

    free(scrambled_password);

    if ((len = SSL_read(ssl, reply, REPLY_LEN)) <= 0)
    {
	ERR_print_errors_fp(stderr);
	exit(1);
    }
    reply[len] = '\0';

    if (strncmp(reply, correct_reply, sizeof(correct_reply)))
    {
	fprintf(stderr, "Authorization failed.\n");
	return -1;
    }
    return 0;
}
/* }}} */


/* create_cvspass_start_of_line() {{{
 * Creates the start of line in .cvspass file we should look for.
 */
char *create_cvspass_start_of_line(char const *cannonical_cvsroot)
{
    int start_of_line_len = strlen(cannonical_cvsroot) + 4;
    char *start_of_line = xmalloc(start_of_line_len + 1);

    /* We create the beginning of line to look for */
    /* TOFIX: we support only the new format of .cvspass, that is:
     * /1 user@host:port/path Aencoded_password
     */
    strcpy(start_of_line, "/1 ");
    strcat(start_of_line, cannonical_cvsroot);
    strcat(start_of_line, " ");
    return start_of_line;
}
/* }}} */


/* save_cvspass_line() {{{
 * Writes one line of .cvspass to the given FILE
 * Returns 0 on success, -1 on error.
 */
int save_cvspass_line(FILE *f, char const *fpath, 
		      char const *start_of_line, char const *scrambled_password)
{
    if (fputs(start_of_line, f) < 0 ||
	fputs(scrambled_password, f) < 0 ||
	fputs("\n", f) < 0)
    {
	fprintf(stderr, "Cannot write to `%s'.\n", fpath);
	return -1;
    }
    return 0;
}
/* }}} */


/* copy_cvspass_file() {{{
 * Will copy cvspass file from one FILE to another. If given start_of_line_to_exclude, it will
 * exclude all lines that start with the string.
 * Returns 0 on success, -1 on error.
 */
int copy_cvspass_file(FILE *from, FILE *to, char const *topath,
		      char const *start_of_line_to_exclude)
{
    /* line buffer ... */
    #define LINE_LEN 1024
    char line[LINE_LEN];
    int start_of_line_len = 0;

    if (start_of_line_to_exclude)
	start_of_line_len = strlen(start_of_line_to_exclude);

    /* We go through the file */
    while (fgets(line, LINE_LEN, from))
    {
	/* If the records match, we skip it. */
	if (start_of_line_to_exclude &&
	    !strncmp(line, start_of_line_to_exclude, start_of_line_len))
	    continue;
	/* Otherwise we output it... */
	if (fputs(line, to) < 0)
	{
	    fprintf(stderr, "Cannot write to file `%s'.\n", topath);
	    return -1;
	}
    }
    return 0;
}
/* }}} */


/* save_password() {{{
 * Saves password to the .cvspass file
 */
void save_password(char *username, char *password, char *server, int port, char *root)
{
    /* files ... */
    char *cvspasspath = construct_cvspass_filename();
    char *tmppath = xstrdup("/tmp/ccvsshXXXXXX");
    FILE *cvsf = NULL, *tmpf = NULL;
    /* password and cannonical CVSROOT ... */
    char *scrambled_password = scramble(password);
    char *cannonical_cvsroot = get_cannonical_cvsroot(username, server, port, root);
    /* line we are looking for ... */
    char *start_of_line = create_cvspass_start_of_line(cannonical_cvsroot);


    /* {{{ We open the .cvspass file. If we don't have source file, we save the file directly into
     * .cvspass */
    if (!(cvsf = fopen(cvspasspath, "r")))
    {
	/* The file doesn't exist yet (TOFIX: or I don't have permissions :)
	 * So we try to open it for writing and save the password directly into it...
	 */
	if (!(cvsf = fopen(cvspasspath, "w")))
	{
	    fprintf(stderr, "Cannot open `%s'.\n", cvspasspath);
	    goto error_exit;
	}
	if (save_cvspass_line(cvsf, cvspasspath, start_of_line, scrambled_password))
	    goto error_exit;
	fclose(cvsf);
	return;
    }
    /* }}} */

    /* We open temporary file {{{ */
    if (!(tmpf = fdopen(mkstemp(tmppath), "w")))
    {
	fprintf(stderr, "Cannot open temporary file `%s'.\n", tmppath);
	goto error_exit;
    }
    /* }}} */
    
    /* We copy the file except the modified record. {{{ */
    if (copy_cvspass_file(cvsf, tmpf, tmppath, start_of_line))
	goto error_exit;
    /* }}} */

    /* And at the end we append the scrambled password {{{ */
    if (save_cvspass_line(tmpf, tmppath, start_of_line, scrambled_password))
	goto error_exit;
    /* }}} */

    /* let's reopen && rewind the files... {{{ */
    if (!(tmpf = freopen(tmppath, "r", tmpf)))
    {
	fprintf(stderr, "Cannot reopen temporary file `%s' for reading.\n", tmppath);
	goto error_exit;
    }
    if (!(cvsf = freopen(cvspasspath, "w", cvsf)))
    {
	fprintf(stderr, "Cannot open `%s' for writing.\n", cvspasspath);
	goto error_exit;
    }
    /* }}} */

    /* we copy the stuff back to the original file {{{ */
    if (copy_cvspass_file(tmpf, cvsf, cvspasspath, NULL))
	goto error_exit;
    /* }}} */

error_exit:
    unlink(tmppath);
    free(cvspasspath);
    free(tmppath);
    if (cvsf)
	fclose(cvsf);
    if (tmpf)
	fclose(tmpf);
}
/* }}} */


/* delete_pasword() {{{
 * Delete password from the .cvspass file
 */
void delete_password(char *username, char *server, int port, char *root)
{
    /* files ... */
    char *cvspasspath = construct_cvspass_filename();
    char *tmppath = xstrdup("/tmp/ccvsshXXXXXX");
    FILE *cvsf = NULL, *tmpf = NULL;
    /* password and cannonical CVSROOT ... */
    char *cannonical_cvsroot = get_cannonical_cvsroot(username, server, port, root);
    /* line we are looking for ... */
    char *start_of_line = create_cvspass_start_of_line(cannonical_cvsroot);

    /* {{{ We open the .cvspass file. */
    if (!(cvsf = fopen(cvspasspath, "r")))
    {
	fprintf(stderr, "Cannot open `%s'.\n", cvspasspath);
	goto error_exit;
    }
    /* }}} */

    /* We open temporary file {{{ */
    if (!(tmpf = fdopen(mkstemp(tmppath), "w")))
    {
	fprintf(stderr, "Cannot open temporary file `%s'.\n", tmppath);
	goto error_exit;
    }
    /* }}} */
    
    /* We copy the file except the deleted record. {{{ */
    if (copy_cvspass_file(cvsf, tmpf, tmppath, start_of_line))
	goto error_exit;
    /* }}} */

    /* let's reopen && rewind the files... {{{ */
    if (!(tmpf = freopen(tmppath, "r", tmpf)))
    {
	fprintf(stderr, "Cannot reopen temporary file `%s' for reading.\n", tmppath);
	goto error_exit;
    }
    if (!(cvsf = freopen(cvspasspath, "w", cvsf)))
    {
	fprintf(stderr, "Cannot open `%s' for writing.\n", cvspasspath);
	goto error_exit;
    }
    /* }}} */

    /* we copy the stuff back to the original file {{{ */
    if (copy_cvspass_file(tmpf, cvsf, cvspasspath, NULL))
	goto error_exit;
    /* }}} */

error_exit:
    unlink(tmppath);
    free(cvspasspath);
    free(tmppath);
    if (cvsf)
	fclose(cvsf);
    if (tmpf)
	fclose(tmpf);
}
/* }}} */


/* get_password() {{{
 * Reads and returns password from the .cvspass file
 * Allocates and returns new string.
 */
char *get_password(char *username, char *server, int port, char *root)
{
    /* files ... */
    char *cvspasspath = construct_cvspass_filename();
    FILE *cvsf = NULL;
    /* password and cannonical CVSROOT ... */
    char *password = NULL;
    char *cannonical_cvsroot = get_cannonical_cvsroot(username, server, port, root);
    /* line we are looking for ... */
    char *start_of_line = create_cvspass_start_of_line(cannonical_cvsroot);
    int start_of_line_len = strlen(start_of_line);
    /* line buffer ... */
    char line[LINE_LEN];

    /* We open the file {{{ */
    if (!(cvsf = fopen(cvspasspath, "r")))
    {
	fprintf(stderr, "Cannot open `%s'.\n", cvspasspath);
	exit(1);
    }
    /* }}} */
    
    /* We go through the file {{{ */
    while (fgets(line, LINE_LEN, cvsf))
    {
	/* If the records match, we skip it. */
	if (!strncmp(line, start_of_line, start_of_line_len))
	{
	    char *c;
	    password = line + start_of_line_len + 1;
	    /* strip new line */
	    if ((c = strchr(password, '\n')))
		*c = '\0';
	    /* and descramble */
	    c = scramble(password);
	    /* after descrambling we have to remove the first A. don't ask me why, I don't wanna
	     * know :/ */
	    password = xstrdup(c+1);
	    free(c);
	    break;
	}
    }
    /* }}} */

    fclose(cvsf);

    free(cvspasspath);

    if (!password)
    {
	fprintf(stderr, "Password for `%s' not found!\n", cannonical_cvsroot);
	exit(1);
    }

    return password;
}
/* }}} */
/* }}} */


/* usage() {{{ */
void usage(char *argv[])
{
    char *basename = strrchr(argv[0], '/');
    if (basename)
	basename++;
    else
	basename = argv[0];

    fprintf(stderr, "Usage: %s login    - login to CVS server\n", basename);
    fprintf(stderr, "       %s logout   - logout from CVS server\n\n", basename);
    fprintf(stderr, "Other than that, this program should be called by cvs.\n");
    fprintf(stderr, "For that, set CVSROOT=:ext:... and CVS_RSH=%s\n\n", basename);
    exit(1);
}
/* }}} */

/* login() {{{ */
void login()
{
    /* yes, it is login. I'll try to separate the info {{{ */
    char *username, *password, *server, *root;
    int port;
    if (extract_cvsroot_info(&username, &password, &server, &port, &root))
	exit(1);
    /* }}} */

    /* if password was not given, ask for it... {{{ */
    if (!password)
	password = getpass("Enter password: ");
    /* }}} */

    /* ... request password authorization and save it ... {{{ */
    init_ssl();
    open_ssl_connection(server, port);
    if (validate_password_on_cvs_server(username, password, root) == 0)
    {
	save_password(username, password, server, port, root);
	puts("Login ok.");
    }
    close_ssl_connection();
    shut_ssl();
    /* }}} */
}
/* }}} */

/* logout {{{ */
void logout()
{
    /* yes, it is logout. I'll try to separate the info {{{ */
    char *username, *password, *server, *root;
    int port;
    if (extract_cvsroot_info(&username, &password, &server, &port, &root))
	exit(1);
    /* }}} */

    /* delete the password from cvspass {{{*/
    delete_password(username, server, port, root);
    puts("Logout ok.");
    /* }}} */
}
/* }}} */

/* pipe_traffic() {{{ */
void pipe_traffic(char *argv[])
{
    char *username, *password, *server, root[LINE_LEN+1];
    int port;
    char *c, *d;
    int pos;
    int l_param_pos;

    /* scan through the argv and find where username and server arguments
     * are located {{{
     */
    for (l_param_pos = 1; l_param_pos < 5; l_param_pos++)
    {
	if (!strncmp(argv[l_param_pos], "-l", 2))
	    break;
    }

    /* if no login specification has been found, this is a problem {{{ */
    if (l_param_pos > 3)
	usage(argv);
    /* }}} */
    /* }}} */

    /* set the server and username based on the position of the -l parameter {{{ */
    if (l_param_pos == 1)
	server = argv[3];
    else
	server = argv[1];
    username = argv[l_param_pos + 1];
    port = SCVS_PORT;
    /* }}} */

    /* root should be on the first line {{{
     * format is:
     * Root /home/cvs
     */
    /* Read a line... {{{ */
    pos = 0;
    while (pos < LINE_LEN
	    && read(0, &root[pos], 1)
	    && root[pos] != '\n')
	pos++;
    root[pos] = '\0';
    /* }}} */
    /* And extract root {{{ */
    for (c = root + 5, d = root; *c && !isspace(*c);)
	*d++ = *c++;
    *d = '\0';
    /* }}} */
    /* }}} */

    /* we read file from .cvspass or die {{{ */
    password = get_password(username, server, port, root);
    /* }}} */

    /* And now we do the actual forwarding... {{{ */
    init_ssl();
    open_ssl_connection(server, port);
    if (validate_password_on_cvs_server(username, password, root) == 0)
    {
	/* First we send the modified root information */
	send_via_ssl("Root ", 0);
	send_via_ssl(root, 0);
	send_via_ssl("\n", 0);
	/* ... and now we just forward the traffic */
	forward_stdinout_to_ssl_connection();
    }
    close_ssl_connection();
    shut_ssl();
    /* }}} */
}
/* }}} */

/* main() {{{ */
int main(int argc, char *argv[])
{
    /* is it login? {{{
     * that is, did I get the following command:
     * 	ccvssh login
     */
    if (argc == 2 && !strcmp(argv[1], "login"))
    {
	login();
	return 0;
    }
    /* }}} */

    /* or logout? {{{
     * that is, did I get the following command:
     * 	ccvssh logout
     */
    if (argc == 2 && !strcmp(argv[1], "logout"))
    {
	logout();
	return 0;
    }
    /* }}} */

    /* otherwise, we're run from cvs client. {{{
     * argv should contain:
     *  { 0: "ccvssh", 1: "cvs.server.cz", 2: "-l", 3: "username", 4: "cvs server" }
     *  NOTE: the order can change - it depends on the -l parameter
     */
    if (argc == 5 && !strcmp(argv[4], "cvs server"))
    {
	pipe_traffic(argv);
	return 0;
    }
    /* }}} */

    /* Unknown combination of parameters */
    usage(argv);
    return 1;
}
/* }}} */
